module CushionDefaults
  # Slight expansion of @Hash@.
  #
  # Records a link to @owner@ and will traverse up the chain of owners looking for a default value for @x@ if no default
  # value for @x@ is specified.
  #
  # In general, a program should treat this class like a normal Hash, with two salient differences:
  # - The {DefaultsHash#+} method has been overridden.
  # - The {DefaultsHash#has_ish_key?}, which returns true if this or a parent has the key specified.
  class DefaultsHash < Hash
    # @default_proc@ does most of the work.

    # Must be initialized with an @owner@ class.

    # Set of defaults marked as pushy.
    attr_reader :pushy_defaults
    # Set of defaults marked as polite.
    attr_reader :polite_defaults
    # Set of defaults frozen
    attr_reader :frozen_defaults

    # @param owner [Class] class for which this DefaultsHash holds the defaults.
    def initialize(owner)
      self.default_proc = proc { |_hash, key| super_defaults[key] }
      # look in the superclass' @defaults@ hash, if a key is not found in this class' @defaults@ hash
      @owner = owner
      @pushy_defaults = Set.new
      @polite_defaults = Set.new
      @frozen_defaults = Set.new
    end

    # @!group Custom/New Hash Methods

    # Allows addition of hashes. Works as expected.
    #
    # Note that this also enables +=.
    #
    # If @key@ is in @self@ and @y@, then the value of @y@ will replace that in @self@ in the resulting @Hash@.
    #
    # @note This method was incorrectly programmed to modify self in place. In a future release, this will be changed,
    #   and the current behavior should not be depended upon.
    #
    # @param y [Hash] hash whose key/values are to be copied over
    # @return [DefaultsHash] self
    def +(y)
      y.each do |key, val|
        if has_key? key
          CushionDefaults.log("As #{@owner} already has the default #{key} defined, this call to + is overwriting it.", :info)
        end
        self[key] = val
      end
      self
    end

    # Custom key/value set method. Prevents writing a default when a parent has it marked as pushy (and it is not
    # otherwise marked as polite), and it also tells @owner to add methods as needed (if writers or readers are to be
    # automatically added).
    #
    # @raise [FrozenDefaultError] if key is frozen
    def []=(key,val)
      key = key.to_sym
      if !@polite_defaults.include?(key) && pushy_in_parent?(key)
        raise ArgumentError, 'You cannot set a default value marked as pushy in a parent class without first marking it as polite.'
      end
      if has_key?(key) && default_frozen?(key)
        raise FrozenDefaultError.new(@owner, key), "#{@owner}.defaults[:#{key} is frozen!"
      end
      unless has_ish_key?(key)
        add_methods_as_needed(key)
      end
      super(key, val)
    end

    # Custom delete method. We need to check whether or not we should delete the methods associated with the key, and we
    # need to remove the key from @pushy_defaults and @polite_defaults if it exists.
    def delete(key)
      key = key.to_sym
      return unless has_key? key
      remove_methods_as_needed(key)
      #TODO Need some better way to handle when @pushy_defaults is emptied or cleared. Need to talk to children.
      @owner.send :_cascade_polite_symbol, key if @pushy_defaults.delete? key
      @owner.send :_cascade_pushy_symbol, key if @polite_defaults.delete? key
      @frozen_defaults.delete(key)
      super(key)
      CushionDefaults.log("Default for #{key} in #{@owner} deleted.")
    end

    # Custom clear method. We need to check whether or not we should delete the methods associated with the keys, and we
    # need to wipe @pushy_defaults and @polite_defaults.
    def clear
      keys.each do |key|
        remove_methods_as_needed(key.to_sym)
        @owner.send :_cascade_polite_symbol, key if @pushy_defaults.delete? key
        @owner.send :_cascade_pushy_symbol, key if @pushy_defaults.delete? key
      end
      @frozen_defaults.clear
      super
      CushionDefaults.log("All defaults cleared for #{@owner}.", :info)
    end

    # Determine if this @DefaultsHash@ "ish" has a key.  In other words, whether it or any Hashes up the chain of
    # @super_defaults@ has the key @key@.
    #
    # @see #ish_keys
    def has_ish_key?(key)
      # Obviously, this method gives much better performance than calling @ish_keys.include?(key)@.
      if has_key?(key)
        true
      # super_defaults could theoretically be either a DefaultsHash or a regular Hash
      elsif super_defaults
        if super_defaults.respond_to? :has_ish_key?
          super_defaults.has_ish_key?(key)
        else
          super_defaults.has_key?(key)
        end
      else
        false
      end
    end

    # Returns the keys of this class' @defaults@ as well as those of parent classes (if extant).
    #
    # @see #has_ish_key?
    def ish_keys
      if super_defaults
        # super_defaults could be either a DefaultsHash or a regular Hash
        (keys + (super_defaults.respond_to?(:ish_keys) ? super_defaults.ish_keys : super_defaults.keys)).uniq
      else
        keys
      end
    end

    # @!group Frozen Defaults

    # @return [boolean] true if the current default (not its value) is frozen, false otherwise.
    def default_frozen? key
      key = key.to_sym
      @frozen_defaults.include? key
    end

    # Freezes the specified default (not its value).
    #
    # @see {#thaw_default!}
    # @see {ClassMethods#freeze_default}
    # @see {ClassMethods#deep_freeze_default}
    # @return [Set] @frozen_defaults if key was not already present, nil otherwise
    def freeze_default! key
      @frozen_defaults.add? key
    end

    # Thaws the specified default (not its value).
    #
    # @see {#freeze_default!}
    # @see {ClassMethods#thaw_default}
    # @return [Set] @frozen_defaults if key was present, nil otherwise
    def thaw_default! key
      @frozen_defaults.delete? key
    end

    # @!group Pushy Defaults

    # Mark sym as pushy
    def pushy!(sym)
      CushionDefaults.conf.we_have_a_pushy!
      @pushy_defaults.add(sym)
      @polite_defaults.delete(sym)

      @owner.send :_cascade_pushy_symbol, sym

      CushionDefaults.log(":#{sym} in #{@owner} is now pushy.")
    end

    # Mark sym as not pushy / polite.
    def not_pushy!(sym)
      @pushy_defaults.delete(sym)
      @polite_defaults.add(sym)

      @owner.send :_cascade_polite_symbol, sym

      # we may need now to define a reader for it
      @owner.send :_update_cushion_reader, sym
      CushionDefaults.log("#{sym} in #{@owner} is now polite.")
    end

    # Determine whether this or a parent defaults hash has declared sym to be pushy.
    def pushy?(sym)
      if @polite_defaults.include?(sym)
        # white list has priority
        false
      elsif @pushy_defaults.include?(sym)
        true
      else
        pushy_in_parent?(sym)
      end
    end

    # Determine whether a parent defaults hash has declared sym to be pushy.
    def pushy_in_parent?(sym)
      if super_defaults.respond_to? :pushy?
        super_defaults.pushy?(sym)
      else
        false
      end
    end

    # Inverse of {#pushy?}
    def not_pushy?(sym)
      !pushy?(sym)
    end

    # @!group Debug

    # Returns the nearest class that has a default set for sym, or nil if no default is set for it.
    def where_is_that_default_again(sym)
      if has_key? sym
        @owner
      elsif super_defaults.respond_to? :where_is_that_default_again
        super_defaults.where_is_that_default_again sym
      else
        nil
      end
    end

    # @!endgroup

    # Tell owner to add readers and writers for a newly-added key, as configured.
    # @param key [#to_sym] name of new default
    # @see Configuration#update_readers
    # @see Configuration#update_writers
    # @see ClassMethods#cushion
    def add_methods_as_needed(key)
      @owner.send :_update_cushion_reader, key
      @owner.cushion_writer key.to_sym if CushionDefaults.conf.update_writers
    end

    # Tell owner to remove methods for a newly-deleted key, as configured.
    # @param key [#to_sym] name of deleted default
    # @see Configuration#update_readers
    # @see Configuration#update_writers
    # @see ClassMethods#remove_reader
    # @see ClassMethods#remove_writer
    def remove_methods_as_needed(key)
      @owner.remove_cushion_reader key.to_sym if CushionDefaults.conf.update_readers
      @owner.remove_bang_reader key.to_sym if CushionDefaults.conf.update_readers
      @owner.remove_cushion_writer key.to_sym if CushionDefaults.conf.update_writers
    end

    protected

    # Return (unless cached) a reference to the superclass' @defaults@ hash.
    def super_defaults
      @super_defaults ||= @owner.superclass.respond_to?(:defaults) ? @owner.superclass.defaults : {}
    end
  end
end